/*
 * Copyright (c) 2016, Freescale Semiconductor, Inc.
 * Copyright 2016-2017 NXP
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of the copyright holder nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "fsl_debug_console.h"
#include "board.h"
#include "fsl_pwm.h"
#include "fsl_cmp.h"

#include "pin_mux.h"
#include "fsl_xbara.h"
/*******************************************************************************
 * Definitions
 ******************************************************************************/
/* The PWM base address */
#define DEMO_PWM_BASEADDR             PWMA
#define DEMO_PWM_SUBMODULE            kPWM_Module_0
#define DEMO_PWM_FAULT_INPUT_PIN      kPWM_Fault_0
#define DEMO_PWM_CONTROL_SUBMODULE    kPWM_Control_Module_0
#define DEMO_PWM_CHANNEL              kPWM_PwmA
#define DEMO_PWM_DELAY_VAL            0x0FFFU

#define DEMO_PWM_CHANNEL_LOCATION_ON_BOARD   "J501 Pin 9"

#define PWM_SOURCE_CLOCK_IN_HZ CLOCK_GetFreq(kCLOCK_FastPeriphClk)

#define DEMO_CMP_BASE                   CMP0
#define DEMO_CMP_USER_CHANNEL           2U
#define DEMO_CMP_DAC_CHANNEL            7U
#define DEMO_CMP_INPUT_PIN_LOCATION_ON_BOARD "J501_13"

#define DEMO_DEADTIME_VAL 650U
#define PWM_SRC_CLK_FREQ CLOCK_GetFreq(kCLOCK_FastPeriphClk)

/*******************************************************************************
 * Prototypes
 ******************************************************************************/
/* Sets up the PWM signals for a PWM submodule */
static void PWM_InitPhasePwm(void);

/* Sets up the PWM fault protection */
static void PWM_SetupFaultPwm(void);

/*******************************************************************************
 * Variables
 ******************************************************************************/

/*******************************************************************************
 * Code
 ******************************************************************************/

static void PWM_InitPhasePwm(void)
{
    /* Structure of setup PWM */
    pwm_signal_param_t pwmSignal;
    uint16_t deadTimeVal;
    uint32_t pwmSourceClockInHz, pwmFrequencyInHz = 1U;

    pwmSourceClockInHz = PWM_SRC_CLK_FREQ;
    /* Set deadtime count */
    deadTimeVal = ((uint64_t)pwmSourceClockInHz * DEMO_DEADTIME_VAL) / 1000000000U;
    pwmSignal.pwmChannel = DEMO_PWM_CHANNEL;
    pwmSignal.level = kPWM_HighTrue;
    pwmSignal.dutyCyclePercent = 50U; /* 50 percent dutycycle */
    pwmSignal.deadtimeValue = deadTimeVal;

    PWM_SetupPwm(DEMO_PWM_BASEADDR, DEMO_PWM_SUBMODULE, &pwmSignal, 1U, kPWM_SignedCenterAligned,\
                             pwmFrequencyInHz,pwmSourceClockInHz);

    /* Set the load okay bit for all submodules to load registers from their buffer */
    PWM_SetPwmLdok(DEMO_PWM_BASEADDR, DEMO_PWM_CONTROL_SUBMODULE, true);
}

static void PWM_SetupFaultPwm(void)
{
    /* Structure of the parameters to configure a PWM fault  */
    pwm_fault_param_t pwmFaultParam;

    /* Setup Fault config */
    /* No combination path is available */
    pwmFaultParam.enableCombinationalPath = false;

    /* Logic 1 on the fault input pin indicates fault */
    pwmFaultParam.faultLevel = true;
    /*
     * Automatic fault clearing
     * If use Manual fault clearing mode, then the user must clear fault flags
     */
    pwmFaultParam.faultClearingMode = kPWM_Automatic;
    pwmFaultParam.recoverMode = kPWM_RecoverFullCycle;

    PWM_SetupFaults(DEMO_PWM_BASEADDR, DEMO_PWM_FAULT_INPUT_PIN, &pwmFaultParam);

    DEMO_PWM_BASEADDR->SM[DEMO_PWM_SUBMODULE].DISMAP[DEMO_PWM_SUBMODULE] = 0x00;
    DEMO_PWM_BASEADDR->SM[DEMO_PWM_SUBMODULE].DISMAP[DEMO_PWM_SUBMODULE] |= PWM_DISMAP_DIS0A(1);
}

int main(void)
{
    /* Structure of initialize PWM */
    pwm_config_t pwmConfig;

    cmp_config_t mCmpConfigStruct;
    cmp_dac_config_t mCmpDacConfigStruct;

    uint8_t ret = 0U;
    uint16_t i = 0U;
    uint32_t pwmVal = 4U;

    BOARD_InitPins();
    BOARD_BootClockRUN();
    BOARD_InitDebugConsole();

    /* Set the PWM Fault inputs to CMP0 output pin */
    XBARA_Init(XBARA);
    XBARA_SetSignalsConnection(XBARA, kXBARA_InputCmp0Output, kXBARA_OutputPwmAFault0);

    /*
     * mCmpConfigStruct.enableCmp = true;
     * mCmpConfigStruct.hysteresisMode = kCMP_HysteresisLevel0;
     * mCmpConfigStruct.enableHighSpeed = false;
     * mCmpConfigStruct.enableInvertOutput = false;
     * mCmpConfigStruct.useUnfilteredOutput = false;
     * mCmpConfigStruct.enablePinOut = false;
     * mCmpConfigStruct.enableTriggerMode = false;
     */
    CMP_GetDefaultConfig(&mCmpConfigStruct);
    /* Init the CMP comparator. */
    CMP_Init(DEMO_CMP_BASE, &mCmpConfigStruct);

    /* Configure the DAC channel. */
    mCmpDacConfigStruct.referenceVoltageSource = kCMP_VrefSourceVin2; /* VCC. */
    mCmpDacConfigStruct.DACValue = 32U;                               /* Half voltage of logic high level. */
    CMP_SetDACConfig(DEMO_CMP_BASE, &mCmpDacConfigStruct);
    CMP_SetInputChannels(DEMO_CMP_BASE, DEMO_CMP_USER_CHANNEL, DEMO_CMP_DAC_CHANNEL);

    /*
     * pwmConfig.enableDebugMode = false;
     * pwmConfig.enableWait = false;
     * pwmConfig.reloadSelect = kPWM_LocalReload;
     * pwmConfig.faultFilterCount = 0;
     * pwmConfig.faultFilterPeriod = 0;
     * pwmConfig.clockSource = kPWM_BusClock;
     * pwmConfig.prescale = kPWM_Prescale_Divide_1;
     * pwmConfig.initializationControl = kPWM_Initialize_LocalSync;
     * pwmConfig.forceTrigger = kPWM_Force_Local;
     * pwmConfig.reloadFrequency = kPWM_LoadEveryOportunity;
     * pwmConfig.reloadLogic = kPWM_ReloadImmediate;
     * pwmConfig.pairOperation = kPWM_Independent;
     */
    PWM_GetDefaultConfig(&pwmConfig);
    /* Fault filter count */
    pwmConfig.faultFilterCount = 0x07U;
    /* Fault filter period; value of 0 will bypass the filter */
    pwmConfig.faultFilterPeriod = 0x14U;
    pwmConfig.prescale = kPWM_Prescale_Divide_1;
    /* Use full cycle reload */
    pwmConfig.reloadLogic = kPWM_ReloadPwmFullCycle;
    /* PWM A & PWM B operate as 2 independent channels */
    pwmConfig.pairOperation = kPWM_Independent;
    pwmConfig.enableDebugMode = true;

    /* Initialize submodule 0 */
    ret = PWM_Init(DEMO_PWM_BASEADDR, DEMO_PWM_SUBMODULE, &pwmConfig);
    if(ret != kStatus_Success)
    {
        PRINTF("\r\nPWM INIT FAILED");
        return 1;
    }

    PWM_InitPhasePwm();

    PWM_SetupFaultPwm();

    PRINTF("\r\n\r\nWelcome to PWM Fault demo");
    PRINTF("\r\nUse oscilloscope to see PWM signal at probe pin: %s", DEMO_PWM_CHANNEL_LOCATION_ON_BOARD);
    PRINTF("\r\nConnect pin %s to high level and ground to see change.", DEMO_CMP_INPUT_PIN_LOCATION_ON_BOARD);

    PWM_StartTimer(DEMO_PWM_BASEADDR, DEMO_PWM_CONTROL_SUBMODULE);
    CMP_Enable(DEMO_CMP_BASE, true);

    while(1)
    {
        /* Time delay for update the duty cycle percentage */
        for (i = 0U; i < DEMO_PWM_DELAY_VAL; i++)
        {
            __NOP();
        }
        pwmVal = pwmVal + 4;

        /* Reset the duty cycle percentage */
        if (pwmVal > 100)
        {
            pwmVal = 4;
        }

        /* Update duty cycles for PWM signals */
        PWM_UpdatePwmDutycycle(DEMO_PWM_BASEADDR, DEMO_PWM_SUBMODULE, DEMO_PWM_CHANNEL, kPWM_SignedCenterAligned, pwmVal);
        /* Set the load okay bit for all submodules to load registers from their buffer */
        PWM_SetPwmLdok(DEMO_PWM_BASEADDR, DEMO_PWM_CONTROL_SUBMODULE, true);
    }
}
